import json
from . import fields
from . import validate

try:
    import urllib2 as request
except ImportError:
    from urllib import request

ENABLED_HUMAN_VALUE_MAP = {
    0 : 'Disabled',
    1 : 'Enabled'
}

class Thermostat(object):
    """
    This class implements the most basic functionality of communicating with
    an actual thermostat.
    """
    MODEL = ''
    # The current API doesn't require this header, but it also doesn't hurt,
    # and it's the right thing to do.
    JSON_HEADER = {'Content-Type' : 'application/json'}

    def __init__(self, host):
        self.host = host

    def get(self, relative_url):
        """
        :param relative_url:    The relative URL from the root of the website.

        :returns:   file-like object as returned by urllib[2,.request].urlopen
        """
        url = self._construct_url(relative_url)
        return request.urlopen(url)

    def post(self, relative_url, value):
        """
        :param relative_url:    The relative URL from the root of the website.
        :param value:           Value to set this attribute to

        :returns:   file-like object as returned by urllib[2,.request].urlopen
        """
        url = self._construct_url(relative_url)
        request_instance = request.Request(url, value, self.JSON_HEADER)
        return request.urlopen(request_instance)

    def _construct_url(self, relative_url):
        """
        :param relative_url:    The relative URL from the root of the website

        :returns:   Full URL, for example 'http://192.168.0.2/tstat'
        """
        return 'http://%s/%s' % (self.host, relative_url.lstrip('/'))


class CommonThermostat(Thermostat):
    """
    This class implements the common API features that are available and work
    across all models of thermostat.
    """
    def reboot(self):
        """reboots the thermostat"""
        response = self.post('/sys/command', json.dumps({'command' : 'reboot'}).encode('utf-8'))
        validate.validate_response(response)

    ### tstat subsystem ###
    tstat = fields.ReadOnlyField('/tstat', None)
    model = fields.ReadOnlyField('/tstat/model', 'model')
    version = fields.Field('/tstat/version', 'version')
    temp = fields.ReadOnlyField('/tstat', 'temp')
    tmode = fields.Field('/tstat', 'tmode',
        human_value_map={
            0 : 'Off',
            1 : 'Heat',
            2 : 'Cool',
            3 : 'Auto'
    })
    fmode = fields.Field('/tstat', 'fmode',
        human_value_map={
            0 : 'Auto',
            1 : 'Auto/Circulate',
            2 : 'On'
    })
    override = fields.ReadOnlyField('/tstat', 'override',
        human_value_map=ENABLED_HUMAN_VALUE_MAP)
    hold = fields.Field('/tstat', 'hold', human_value_map=ENABLED_HUMAN_VALUE_MAP)
    t_heat = fields.Field('/tstat/ttemp', 't_heat', post_url='/tstat')
    t_cool = fields.Field('/tstat/ttemp', 't_cool', post_url='/tstat')
    tstate = fields.ReadOnlyField('/tstat', 'tstate',
        human_value_map={
            0 : 'Off',
            1 : 'Heat',
            2 : 'Cool'
    })
    fstate = fields.ReadOnlyField('/tstat', 'fstate',
        human_value_map={
            0 : 'Off',
            1 : 'On'
    })
    time = fields.Field('/tstat', 'time')
    pump = fields.ReadOnlyField('/tstat/hvac_settings', 'pump',
        human_value_map={
            1 : 'Normal',
            2 : 'Heat Pump'
    })
    aux_type = fields.ReadOnlyField('/tstat/hvac_settings', 'aux_type',
        human_value_map={
            1 : 'Gas',
            2 : 'Electric'
    })
    hvac_code = fields.ReadOnlyField('/tstat/hvac_settings', 'hvac_code',
        human_value_map={
            1 : '1 stage heat, 1 stage cool',
            2 : '2 stage heat, 1 stage cool',
            3 : '2 stage heat, 2 stage cool',
            4 : '2 stage heat, 1 stage cool',
            5 : '2 stage heat, 2 stage cool',
            10 : '1 stage pump, 1 stage aux',
            11 : '1 stage pump, 1 stage aux',
            12 : '1 stage pump, no aux',
    })
    # This isn't documented. It might be postable, but I'm not going to try.
    power = fields.ReadOnlyField('/tstat/power', 'power')

    program_cool = fields.ReadOnlyField('/tstat/program/cool', None)
    program_heat = fields.ReadOnlyField('/tstat/program/heat', None)
    datalog = fields.ReadOnlyField('/tstat/datalog', None)

    ### sys subsystem ###
    sys = fields.ReadOnlyField('/sys', None)
    name = fields.Field('/sys/name', 'name')
    services = fields.ReadOnlyField('/sys/services', None)
    mode = fields.Field('/sys/mode', 'mode', human_value_map={
        0 : 'Provisioning',
        1: 'Normal'
    })
    network = fields.ReadOnlyField('/sys/network', None)
    security = fields.ReadOnlyField('/sys/network', 'security',
        human_value_map = {
            1 : 'WEP',
            3 : 'WPA',
            4 : 'WPA2 Personal'
    })

    ### cloud subsystem ###
    cloud = fields.ReadOnlyField('/cloud', None)

    ### methods ###
    def set_day_program(self, heat_cool, day, program):
        """
        Sets the program for a particular day. See the API docs for details,
        as it is a bit complicated.

        :param heat_cool:   Ether the string 'heat' or 'cool'
        :param day:         One of 'mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'
        :param program:     See thermostat API docs
        :type program:      dict
        """
        self.post('/tstat/program/%s/%s' % (heat_cool, day), json.dumps(program).encode('utf-8'))


class CT50v194(CommonThermostat):
    """
    Defines API features that differ for this specific model from
    CommonThermostat
    """
    MODEL = 'CT50 V1.94'




import json

def validate_response(response, content=None):
    """
    raises AttributeError if the HTTP response code is not 200, or if the value
    returned by the server indicates an error.

    :param response:    the response returned by a call to urllib.request.urlopen
    :type response:     file-like object
    :param content:     the JSON-deserialized value returned by the server.
                        This is required if the response was "read" prior to
                        calling this method.
    :type content:      any iterable collection 
    :returns:           None
    :raises:            AttributeError
    """
    if response.getcode() != 200:
        raise AttributeError('HTTP code %d. %s' % (response.code, response.msg))

    if content is None:
        content = json.loads(response.read().decode('utf-8'))

    for error_field in ('error_msg', 'error'):
        if error_field in content:
            raise AttributeError('Error message from thermostat: %s' % content[error_field])



from .thermostat import Thermostat, CommonThermostat, CT50v194
from . import discover
from . import fields

THERMOSTATS = (CT50v194,)

def get_thermostat_class(model):
    """
    :param model:   string representation of the thermostat's model, in
                    whatever format the thermostat itself returns.
    :type model:    str

    :returns:       subclass of CommonThermostat, or None if there is not a
                    matching subclass found in THERMOSTATS.
    """
    for thermostat in THERMOSTATS:
        if issubclass(thermostat, Thermostat) and thermostat.MODEL == model:
            return thermostat

def get_thermostat(host_address=None):
    """
    If a host_address is not passed, auto-discovery will happen. Auto-discovery
    will only succeed then exactly 1 thermostat is on your network.

    :param host_address:    optional address for a thermostat. This can be an
                            IP address or domain name.

    :returns:   instance of a CommonThermostat subclass, or None if a matching
                subclass cannot be found.
    """
    if host_address is None:
        host_address = discover.discover_address()
    initial = CommonThermostat(host_address)
    model = initial.model.get('raw')
    thermostat_class = get_thermostat_class(model)
    if thermostat_class is not None:
        return thermostat_class(host_address)



import json

from . import validate

class Field(object):
    """
    Instances of this class act as descriptors on the Thermostat class. They
    define a piece of data from the API that can be accessed with GET and
    optionally with POST.
    """
    def __init__(self, url, name, human_value_map=None, post_url=None):
        """
        :param url:             relative URL to use for GET and POST, except
                                when post_url is defined.
        :param name:            key value for the data point you want. Most
                                responses come as a dictionary, for example
                                {'temp' : 73}, so you would pass 'temp' for
                                this argument.
        :param human_value_map: An optional dictionary where keys are actual
                                values the server might return, and values
                                are a human-readable form. This is useful for
                                example when a return value of "0" actually
                                means "Off". If you pass this argument, make
                                sure it includes all possible values.
        :param post_url:        A few items require you to POST to a different
                                URL than where you GET. Use this argument to
                                specify a POST URL, and the 'url' argument will
                                continue to be used for GETs.
        """
        self.url = url
        self.name = name
        self.human_value_map = human_value_map
        self.post_url = post_url

    def __get__(self, instance, owner):
        response = instance.get(self.url)
        envelope = json.loads(response.read().decode('utf-8'))
        validate.validate_response(response, envelope)
        return self._build_get_return(envelope)

    def _build_get_return(self, envelope):
        """
        :param envelope:    raw value returned from the thermostat, which
                            usually is a dict. There are some attributes that
                            don't come with an envelope, and are just the raw
                            value by itself. In that case, set name=None.

        :returns:   dict with key 'raw' whose value is the value this Field is
                    setup to return. For example, it might be the current temp.
                    If there is a human_value_map, this dict will have an
                    additional 'human' key whose value is a human-readable
                    version of the raw value.
        """
        # Some URLs, like /tstat, don't have an envelope
        ret = {'raw' : envelope.get(self.name) if self.name else envelope}
        if self.human_value_map:
            ret['human'] = self._convert_to_human(ret['raw'])
        return ret

    def __set__(self, instance, value):
        data = json.dumps({self.name: value}).encode('utf-8')
        response = instance.post(self.post_url or self.url, data)
        validate.validate_response(response)

    def _convert_to_human(self, value):
        """
        :param value:   raw value retrieved from the thermostat and removed
                        from its envelope

        :returns:   human-readable version of the value, as retrieved from
                    self.human_value_map
        """
        try:
            return self.human_value_map[value]
        except KeyError:
            raise AttributeError('Human readable value not known for raw value %s' %
                (str(value)))


class ReadOnlyField(Field):
    """For read-only values like the current temperature"""
    def __set__(self, instance, value):
        raise TypeError('This attribute does not support writes.')



import socket
import struct

IP_ADDRESS = '239.255.255.250'
PORT = 1900
MESSAGE = """TYPE: WM-DISCOVER\r\nVERSION: 1.0\r\n\r\nservices: com.marvell.wm.system*\r\n\r\n""".encode('utf-8')

def discover_address():
    """
    The example discovery program provided by Radio Thermostat sets the IP packet's
    TTL to 3. I'm not sure why that would be a good idea, because the default value
    of 1 (in the case of multicast) seems reasonable. I'm content to say that this
    method will discover a thermostat on your local network ONLY, and in any more
    complicated case, you should supply a FQDN or IP address.
    """

    with ExitingSocket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP) as sock:
        # set the receive timeout to 1 second
        sock.settimeout(1)
        # make the address reuseable
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.sendto(MESSAGE, (IP_ADDRESS, PORT))

        ip_mreq = struct.pack("=4sl", socket.inet_aton(IP_ADDRESS), socket.INADDR_ANY)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, ip_mreq)
        thermostats = []

        while True:
            try:
                data, thermostat = sock.recvfrom(4096)
                # append IP address only
                thermostats.append(thermostat[0])
            except socket.timeout:
                break

        if len(thermostats) == 0:
            raise IOError('No thermostats were found')
        if len(thermostats) > 1:
            raise IOError("Found %d thermostats and I don't know which to pick." % len(thermostats))
        return thermostats[0]

class ExitingSocket(socket.socket):
    """
    This is a socket subclass that can be used with the "with" statement. It
    will attempt to clean up and close itself on exit.
    """
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        try:
            self.setsockopt(socket.SOL_IP, socket.IP_DROP_MEMBERSHIP, socket.inet_aton(IP_ADDRESS) + socket.inet_aton('0.0.0.0'))
        except socket.error:
            pass
        self.close()



#!/usr/bin/env python

from distutils.core import setup

long_desc = open('README.rst').read()

setup(name='radiotherm',
    version='1.0',
    description='client library for wifi thermostats sold by radiothermostat.com',
    long_description=long_desc,
    packages=('radiotherm',),
    license='BSD',
    author='Michael Hrivnak',
    author_email='mhrivnak@hrivnak.org',
    url='https://github.com/mhrivnak/radiotherm',
    classifiers=(
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Libraries',
    )
)
